"""Basic functions for replicator dynamics

When using this module from an application typically only the the Function
GetReplicator' will be needed to generate a suitable replicator dynamics
function.
"""

import random, warnings
from Compatibility import *

#-------------------------------------------------------------------------------
#
# Some miscellanous classes, functions etc.
#
#-------------------------------------------------------------------------------


# The following version would only compatible with python version 2.2 and above!

##def GenTuples(dimension, count):
##    """Generates an iterator of number tuples
##    Each number is counted from 0 to count-1 from right to left. Example:
##    For dimension=2, count=3 the generated iterator yields (0,0), (0,1),
##    (0,2), (1,0), (1,1), (1,2), (2,0), (2,1), (2,2)
##
##    dimension - positive integer, Dimension of the generated tuples
##    count - positive integer, Range of each digit of the generated tuples
##    return value - iterator of tuples
##    """
##    if dimension == 1:
##        for i in range(count): yield (i,)
##    elif dimension > 1:
##        for i in range(count):
##            for k in GenTuples(dimension-1, count):
##                yield (i,)+k
##    else:  raise ValueError, "tuple dimension "+dimention+" is not allowed!"


class GenTuples(object):
    """Iterator of number tuples

    Each number is counted from 0 to count-1 from right to left. Example:
    For dimension=2, count=3 the generated iterator yields (0,0), (0,1),
    (0,2), (1,0), (1,1), (1,2), (2,0), (2,1), (2,2)
    """
    def __init__(self, dimension, count):
        """Constructs the GenTuples iterator object.

        dimension - positive integer, Dimension of the generated tuples
        count - positive integer, Range of each digit of the generated tuples
        """
        self.dimension = dimension
        self.count = count
        self.stop = count**dimension
    
    def __getitem__(self, i):
        """Returns the i-th tuple in line."""
        if i >= self.stop:  raise IndexError
        l = []
        for x in xrange(self.dimension):
            n = i % self.count
            l.insert(0, n)
            i /= self.count
        return tuple(l)


##class ArrayWrapper(object):
##    """Wraps a function or a nested list or tuple as a multidimensional array
##    Access to a cell of an ArrayWrapper object, e.g. "aw[a,b,c]" will be
##    mapped to a function call "f(a,b,c)" or to a nested sequence "l[a][b][c]"
##    respectively.
##
##    The  class is syntactical sugar to provide similar access to
##    pay functions or lists as to Numeric or nummarray arrays.
##
##    Warning: This wrapper does not do any error checking at all!
##    """
##    
##    def __init__(self, obj):
##        """obj - function or tuple of list
##        """
##        if callable(obj):
##            self.function = obj
##        elif hasattr(obj, "__getitem__"):
##            self.obj = obj
##            self.function = self.__listwrapper
##        else:
##            raise TypeError, "ArrayWrapper only supports callables, lists "+\
##                             "or tuples, but not %s!" % type(obj)
##                             
##    def __listwrapper(self, *key):
##        value = self.obj
##        for i in key:  value = value[i]
##        return value
##        
##    def __getitem__(self, key):
##        return self.function(*key)


#-------------------------------------------------------------------------------
#
# Function for generating (population) distributions
#
#-------------------------------------------------------------------------------

#
# In the following a distribution is always a list of floating point numbers
# between 0.0 and 1.0 that add up to 1.0. (It partitions the interval [0.0, 1.0]
# into a number of partitions of possibly different size.)

def RandomDistribution(n):
    """Creates a random distribution

    n - positive integer: Number of elements of in the distribution
    
    return value - list of floats: Distribution,
        with 0.0 <= t[i] <= 0 and sum(t) == 1.0
    """
    assert n >= 0, "The number of elements in a distribution "\
                   "must at least be 0, not %i !" % n
    if n == 0: return ()
    r = [abs(float(i % 2) - random.random()) for i in range(n-1)]
    r.sort()
    r.extend([1.0,0.0]) 
    return [r[i]-r[i-1] for i in range(n)]

#    p = [random.random() for x in xrange(n)]
#    S = sum(p)
#    p = map(lambda x: x/S, p)
#    return p


def UniformDistribution(n, noise=0.0):
    """Creates an a uniform distribution
    
    n - positive integer: Number of elements
    noise - float: Noise level (0.0 <= noise <= 1.0)
    return value - tuple of floats: Distribution,
        with 0.0 <= t[i] <=  1.0 and sum(t) == 1.0
    """
    assert n >= 0, "The number of elements in a distribution "\
                   "must at least be 0, not %i !" % n
    if n == 0: return ()
    if noise < 0.0 or noise > 1.0:
        raise ValueError, "noise = %f, must be 0.0 <= noise <= 1.0" % noise
    r = 1.0/n;  rn = r*(1.0-noise)
    p = [random.uniform(rn, r) for i in xrange(n)]
    N = sum(p)
    for i in xrange(n):  p[i] /= N

##    Alternative noise model
##
##    p = [1.0/n for i in range(n)] 
##    if noise > 0.0 and noise <= 1.0:
##        r = RandomPopulation(n)
##        for i in range(n):  p[i] = p[i]*(1.0-noise) + r[i]*noise
##    elif noise != 0.0:  raise ValueError, "0.0 <= noise <= 1.0, not %f" % noise

    return p


def RandomSelect(distribution):
    """Randomly selects an item from a distribution
    The size of the species determines the probability with which it will be
    selected.

    population - tuple of floats: Distribution
    
    return value - integer: index of the selected element (or partition resp.)
    """
    def rselect(slices, r, a, b):
        if b-a <= 1 : return a
        m = (a+b)/2
        if r >= slices[m]:  return rselect(slices, r, m, b)
        else: return rselect(slices, r, a, m)

    if distribution == []:
        raise ValueError, "Cannot select an element from an empty distribution"
    slices = [0.0]
    for p in distribution:  slices.append(p+slices[-1])
    return rselect(slices, random.random(), 0, len(slices))
 

#-------------------------------------------------------------------------------
#
#   Functions to determine the fitness of a population based on
#   2 or more person games between the individuals.
#
#-------------------------------------------------------------------------------

def _Fitness2(population, payoff, e):
    """Determines the relative fitness of species based on a two player game."""
    n = population
    L = len(n)
    Sn = sum(n)
    p = []    
    for i in xrange(L):
        s = 0.0
        for k in xrange(L):
            if i == k:  
                s += payoff[i,k]*(n[i]+e*(Sn-n[i]))
            else: 
                s += payoff[i,k]*(n[k]-e*n[k])
        p.append(s)
    return p

def _SampledFitness2(population, payoff, e, samples):
    """Determine the relative fitness of the species based on a two player game
    taking into accound a correlation of e and using a representative selection 
    of sample matches instead of taking into account every possible combination.
    """
    n = population
    L = len(n)
#    if isinstance(samples, float):
#        samples = int(samples*L*L+0.5)
    Sn = sum(n)
    p = []
    for i in xrange(L):
        m = []
        for k in xrange(L):
            if i == k:  m.append(n[i]+e*(Sn-n[i]))
            else: m.append(n[k]-e*n[k])
        s = 0.0
        for k in xrange(samples):
            s += payoff[i, RandomSelect(m)]
        p.append(s/samples)
    return p    

def _QuickFitness2(population, payoff):
    """Determines the relative fitness of species based on a two player game
    A faster replacement for 'Fitness2' if correlation e = 0.0.
    """
    n = population
    L = len(n)
    p = []
    for i in xrange(L):
        s = 0.0
        for k in xrange(L):
            s += payoff[i,k]*n[k]
        p.append(s)
    return p

def _SampledQuickFitness2(population, payoff, samples):
    """Determine the relative fitness of the species based on a two player game
    assuming that correlation is 0 and using a representative selection of
    sample matches instead of taking into account every possible combination.
    """
    n = population
    L = len(n)
    p = []
    for i in xrange(L):
        s = 0.0
        for k in xrange(samples):
            s += payoff[i, RandomSelect(n)]
        p.append(s/samples)
    return p
    
def _QuickFitness(population, payoff, N):
    """Determines the relative fitness of the species based on a N-player game 
    Correlation factor e is assumed to be 0.0.
    """
    n = population
    L = len(n)
    p = []
    for i in xrange(L):
        s = 0.0
        for k in GenTuples(N-1, L):
            s += payoff[(i,)+k] * reduce(lambda x,y:x*y, map(lambda x:n[x],k))
        p.append(s)
    return p

def _SampledQuickFitness(population, payoff, N, samples):
    """Determine the relative fitness of the species based on an N-player game
    assuming that correlation is 0 and using a representative selection of
    sample matches instead of taking into account every possible combination.
    """
    n = population
    L = len(n)
    p = []
    for i in xrange(L):
        s = 0.0
        for k in xrange(samples):
            t = tuple([i]+[RandomSelect(n) for v in xrange(N-1)])
            s += payoff[t]
        p.append(s/samples)
    return p

def _Fitness(population, payoff, e, N):
    """Determines the relative fitness of species based on an N-player game"""
    n = population
    L = len(n)
    p = []
    I = sets.Set(range(L))
    for i in xrange(L):
        s = 0.0
        for k in GenTuples(N-1, L):
            K = sets.Set(k)
            Sn = sum(map(lambda y:n[y],filter(lambda x: x != i, I-K)))
            d = {}
            for r in K:
                if r == i:  d[r] = n[r] + e*Sn
                else:       d[r] = n[r] - e*n[r]            
            s += payoff[(i,)+k] * reduce(lambda x,y:x*y, map(lambda x:d[x],k))
        p.append(s)
    return p

def _SampledFitness(population, payoff, e, N, samples):
    """Determines the relative fitness of the species based on an N-player game
    taking into accound a correlation of e and using a representative selection 
    of sample matches instead of taking into account every possible combination.
    """
    n = population
    L = len(n)
    Sn = sum(n)    
    p = []
    for i in xrange(L):
        s = 0.0
        m = []
        for k in xrange(L):
            if i == k:  m.append(n[i]+e*(Sn-n[i]))
            else: m.append(n[k]-e*n[k])        
        for k in xrange(samples):
            t = tuple([i]+[RandomSelect(m) for v in xrange(N-1)])
            s += payoff[t]
        p.append(s/samples)
    return p    


def _FunctionSelector(population, N, samples, stdFunction, sampleFunction,
                      baseParams):
    """Fitness function that determines on the fly whether to use
    a sampling algorithm or not.
    """
    L = len(population)
    combinations = L**N
    if samples < 1.0:  samples = int(combinations*samples+0.5)
    if samples == 0: samples = 1
    if samples < combinations:
        return sampleFunction(population, *(baseParams + (samples,)))
    else: return stdFunction(population, * baseParams)
    

class GenFitnessFunction(object):
    def __init__(self, payoff, e=0.0, N=2, samples=1.0):
        """Generate a fitness function for the given parameters.

            A fitness function if a function that takes a population ratio as
            argument and returns the fitnesses of the species of the population.

        payoff - payoff[a,b,c,...] should yield the payoff of player a when
            playing with players a,b,c,... .

        e - correlation factor, float (range: 0.0 to 1.0): The higher the
            correlation factor the more likely it is that individuals interact
            with individuals of the same species instead of a different species.
            e = 0.0 means random interaction. e = 1.0 means interaction is fully
            confined to the same species.

        N - integer: Number of players in one game, N >= 1

        samples - integer or float: number (or percentage) of samples to be taken.
            If samples = 1.0 or if the number of possible combinations is smaller
            than samples no sampling algorithm will be used.

        return value - function (population) -> fitness: Function that
            computes the fitnesses of the species of the population.

            population - tuple of floats: Population ratio. Each number in the
                tuple represents the share the respective species has from the
                whole population.

            fitness (return value) - list of positive floats: List of the relative
                fitnesses of each species.
        """
        self.payoff, self.e, self.N, self.samples = payoff, e, N, samples

        if e < 0.0 or e > 1.0:
            raise ValueError, "e = %f violates 0.0 <= e <= 1.0"%e
        if samples <= 0: raise ValueError, "samples %f <= 0 !"%samples

        if N == 1:
            if e != 0.0:
                warnings.warn("Correlation e != %f is meaningless if N == 1"%e,
                              UserWarning)
            self.stdFunction = lambda p: [payoff[i,] for i in range(len(p))] 
        elif N == 2:
            if e == 0.0:
                self.stdFunction = _QuickFitness2
                self.params = (payoff,)
                self.sampleFunction = _SampledQuickFitness2
            else:
                self.stdFunction = _Fitness2
                self.params = (payoff, e)
                self.sampleFunction = _SampledFitness2
        elif N > 2:
            if e == 0.0:
                self.stdFunction = _QuickFitness
                self.params = (payoff, N)
                self.sampleFunction = _SampledQuickFitness
            else:
                self.stdFunction = _Fitness
                self.params = (payoff, e, N)
                self.sampleFunction = _SampledFitness
        else: raise ValueError, "N must be >= 1, not %i !" % N
        if samples != 1.0 and N > 1:
            self.function = _FunctionSelector
            self.baseParams = self.params
            self.params = (self.N, self.samples, self.stdFunction,
                           self.sampleFunction, self.baseParams)
        else:  self.function = self.stdFunction
        
    def __call__(self, population):
        return self.function(population, *self.params)

        
#-------------------------------------------------------------------------------
#
#   replicator dynamical functions
#
#-------------------------------------------------------------------------------

def _Replicator(population, fitness, noise):
    """Determines the population distribution of the next generation."""
    n = list(population)
    L = len(population)
    f = fitness(population)
    
    for i in xrange(L):  n[i] *= f[i]- random.uniform(0,f[i]*noise)
    N = sum(n)
    if N == 0.0: # <- this should never be true, except for precision errors!
        n = list(population)
        for i in xrange(L):  n[i] *= 1.0 - random.uniform(0, noise)
        N = sum(n)
        if N == 0.0: return population
    for i in xrange(L): n[i] /= N

##    Alternative noise model:
##
##    for i in xrange(L): n[i] *= f[i]
##    N = sum(n) 
##    if N == 0.0:  n = list(population);  N = 1.0
##    m = N*noise
##    r = [abs(float(i % 2) - random.random()) for i in range(L-1)]
##    r.sort()
##    r.extend([1.0,0.0])
##    for i in xrange(L): n[i] = (n[i] - n[i]*noise + m*(r[i]-r[i-1])) / N
        
    return tuple(n)

def _QuickReplicator(population, fitness):
    """Determines the distribution of species in the next generation
    A faster replacement for Replicator if noise = 0.0. 
    """  
    n = list(population)
    L = len(population)
    f = fitness(population)
    for i in xrange(L): n[i] *= f[i]
    N = sum(n)
    if N == 0.0: return population
    for i in xrange(L): n[i] /= N
    return tuple(n)


class GenDynamicsFunction(object):
    def __init__(self, payoff, e=0.0, noise=0.0, N=2, samples=1.0):
        """Generate a replicator dynamics function

        A replicator dynamics function determines the composition of the
        population of the next generation from the composition of the population
        of the previous generation.

        payoff - two dimensional array of floats: payoff[a,b] should yield the 
            payoff of player a when playing against player b.

        e - correlation factor, float (range: 0.0 to 1.0): The higher the
            correlation factor the more likely it is that individuals interact
            with individuals of the same species instead of a different species.
            e = 0.0 means random interaction. e = 1.0 means interaction is fully
            confined to the same species.

        noise - float: randomness of the redistribution between species. A noise
            level of 0.0 means that redistribution is solely based on fittnes. A
            noise level of 1.0 means that it is totally random.  

        N - integer (N >= 1): Number of players in one game, e.g. number of
            individuals that interact at once.

        samples - integer or float: number (or percentage) of samples to be taken.
            If samples = 1.0 or if the number of possible combinations is smaller
            than samples no sampling algorithm will be used.        

        return value - function (population) -> population: A replicator function
            that takes a population distribution as argument and computes the
            population distribution of the next generation.

            population - tuple of floats: Population ratio. Each number in the
                tuple represents the share the respective species has from the
                whole population.
            return value - the same.
    """
        if noise < 0.0 or noise > 1.0:
            raise ValueError,"noise = %f violates 0.0 <= noise <= 1.0"%noise
        self.fitness = GenFitnessFunction(payoff, e, N, samples)
        if noise == 0.0:
            self.function = _QuickReplicator
            self.params = (self.fitness,)
        else:
            self.noise = noise
            self.function = _Replicator
            self.params = (self.fitness, self.noise)
            
    def __call__(self, population):
        return self.function(population, *self.params)



########################################################################
#
#   Tests
#
########################################################################


if __name__ == "__main__":
    import systemTest
    systemTest.DynamicsTest()
    

